/*
* Copyright(c) 1997-2001 Id Software, Inc.
* Copyright(c) 2002 The Quakeforge Project.
* Copyright(c) 2006 Quetoo.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or(at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
* 
* See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "r_local.h"

/*
 
LIGHT SAMPLING
 
*/

vec3_t pointcolor;
vec3_t lightspot;

int RecursiveLightPoint(mnode_t *node, vec3_t start, vec3_t end){
	float front, back, frac;
	int side;
	mplane_t *plane;
	vec3_t mid;
	msurface_t *surf;
	int s, t, ds, dt;
	int i;
	mtexinfo_t *tex;
	byte *lightmap;
	int maps;
	float samp;
	int r;
	
	if(node->contents != -1)
		return -1;  // didn't hit anything
		
	// calculate mid point
	
	// FIXME: optimize for axial
	plane = node->plane;
	front = DotProduct(start, plane->normal) - plane->dist;
	back = DotProduct(end, plane->normal) - plane->dist;
	side = front < 0;
	
	if((back < 0) == side)
		return RecursiveLightPoint(node->children[side], start, end);
		
	frac = front /(front - back);
	mid[0] = start[0] +(end[0] - start[0]) * frac;
	mid[1] = start[1] +(end[1] - start[1]) * frac;
	mid[2] = start[2] +(end[2] - start[2]) * frac;
	if(plane->type < 3)  // axial planes
		mid[plane->type] = plane->dist;
		
	// go down front side
	r = RecursiveLightPoint(node->children[side], start, mid);
	if(r >= 0)  // hit something
		return r;  
		
	if((back < 0) == side)
		return -1;  // didn't hit anuthing
		
	// check for impact on this node
	VectorCopy(mid, lightspot);
	
	surf = r_worldmodel->surfaces + node->firstsurface;
	if(surf->flags&(SURF_DRAWTURB | SURF_DRAWSKY))
		return 1;  // no lightmaps
	
	for(i = 0; i < node->numsurfaces; i++, surf++){
		tex = surf->texinfo;
		
		s = DotProduct(mid, tex->vecs[0]) + tex->vecs[0][3];
		t = DotProduct(mid, tex->vecs[1]) + tex->vecs[1][3];
		if(s < surf->texturemins[0] ||
				t < surf->texturemins[1])
			continue;
			
		ds = s - surf->texturemins[0];
		dt = t - surf->texturemins[1];
		
		if(ds > surf->extents[0] || dt > surf->extents[1])
			continue;
			
		if(!surf->samples)
			return 0;
			
		ds >>= 4;
		dt >>= 4;
		
		lightmap = surf->samples;
		VectorCopy(vec3_origin, pointcolor);
		if(lightmap){
			lightmap += dt *((surf->extents[0] >> 4) + 1) + ds;
			
			for(maps = 0; maps < MAXLIGHTMAPS && surf->styles[maps] != 255; maps++){
				samp = *lightmap / 255;
				
				pointcolor[0] *= samp;
				pointcolor[1] *= samp;
				pointcolor[2] *= samp;
				
				lightmap += ((surf->extents[0] >> 4) + 1) * ((surf->extents[1] >> 4) + 1);
			}
		}
		
		return 1;
	}
	
	// go down back side
	return RecursiveLightPoint(node->children[!side], mid, end);
}

/*
R_LightPoint
*/
void R_LightPoint(vec3_t p, vec3_t color){
	vec3_t end;
	float r;
		
	if(!r_worldmodel->lightdata){
		color[0] = color[1] = color[2] = 1.0;
		return;
	}
	
	end[0] = p[0];
	end[1] = p[1];
	end[2] = p[2] - 2048;
	
	r = RecursiveLightPoint(r_worldmodel->nodes, p, end);
	
	if(r == -1){
		VectorCopy(vec3_origin, color);
	} else {
		VectorCopy(pointcolor, color);
	}
}

unsigned blocklights[1024];  // allow some very large lightmaps

/*
R_BuildLightMap
 
Combine and scale multiple lightmaps into the 8.8 format in blocklights
*/
void R_BuildLightMap(void){
	int smax, tmax;
	int t;
	int i, size;
	byte *lightmap;
	int maps;
	msurface_t *surf;
	
	surf = r_drawsurf.surf;
	
	smax =(surf->extents[0] >> 4) + 1;
	tmax =(surf->extents[1] >> 4) + 1;
	size = smax * tmax;
	
	// clear to no light
	for(i = 0; i < size; i++)
		blocklights[i] = 0;
	
	if(r_fullbright->value || !r_worldmodel->lightdata){
		return;
	}
	
	// add all the lightmaps
	lightmap = surf->samples;
	if(lightmap)
		for(maps = 0; maps < MAXLIGHTMAPS && surf->styles[maps] != 255; maps++){
			for(i = 0; i < size; i++){
				blocklights[i] += lightmap[i] * 255;
			}
			lightmap += size;  // skip to next lightmap
		}
	
	// bound, invert, and shift
	for(i = 0; i < size; i++){
		t =(int)blocklights[i];
		if(t < 0)
			t = 0;
		t =(255 * 256 - t) >>(8 - VID_CBITS);
		
		if(t <(1 << 6))
			t =(1 << 6);
			
		blocklights[i] = t;
	}
}

