/*
* Copyright(c) 1997-2001 Id Software, Inc.
* Copyright(c) 2002 The Quakeforge Project.
* Copyright(c) 2006 Quetoo.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or(at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
* 
* See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/*
* This file contains ALL Linux specific stuff having to do with the
* SDL refresh.  When a port is being made the following functions
* must be implemented by the port:
*
* SWimp_EndFrame
* SWimp_Init
* SWimp_InitGraphics
* SWimp_SetPalette
* SWimp_Shutdown
* SWimp_SwitchFullscreen
*/

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/mman.h>

#include <SDL.h>

#ifdef OPENGL
#include <GL/gl.h>
#include "gl_local.h"
#else
#include "r_local.h"
#endif

#include "keys.h"

#define JOY_AXIS_X			0
#define JOY_AXIS_Y			1
#define JOY_AXIS_Z			2
#define JOY_AXIS_R			3
#define JOY_AXIS_U			4
#define JOY_AXIS_V			5

static SDL_Surface *surface;
#ifndef OPENGL
static unsigned int sdl_palettemode;
#endif

struct {
	int key;
	int down;
}
keyq[64];
int keyq_head = 0;
int keyq_tail = 0;

int config_notify = 0;
int config_notify_width;
int config_notify_height;

#ifdef OPENGL
glwstate_t glw_state;
#endif

// mouse vars
static qboolean mouse_avail;
static int mouse_buttonstate;
static int mouse_oldbuttonstate;
static int mouse_x, mouse_y;
static int old_mouse_x, old_mouse_y;
static int mx, my;
static float old_windowed_mouse;
static qboolean mouse_active;

static cvar_t *_windowed_mouse;
static cvar_t *m_filter;
static cvar_t *in_mouse;

static qboolean	mlooking;

// state struct passed in Init
static in_state_t *in_state;

static cvar_t *sensitivity;
static cvar_t *lookstrafe;
static cvar_t *m_side;
static cvar_t *m_yaw;
static cvar_t *m_pitch;
static cvar_t *m_forward;
static cvar_t *freelook;

static void Force_CenterView_f(void){
	in_state->viewangles[PITCH] = 0;
}

static void REF_IN_MLookDown(void){
	mlooking = true;
}

static void REF_IN_MLookUp(void){
	mlooking = false;
	in_state->IN_CenterView_fp();
}

void REF_IN_Init(in_state_t *in_state_p){
	
	in_state = in_state_p;
	
	in_mouse = ri.Cvar_Get("in_mouse", "1", CVAR_ARCHIVE);
	_windowed_mouse = ri.Cvar_Get("_windowed_mouse", "1", CVAR_ARCHIVE);
	freelook = ri.Cvar_Get("freelook", "0", 0);
	lookstrafe = ri.Cvar_Get("lookstrafe", "0", 0);
	sensitivity = ri.Cvar_Get("sensitivity", "3", 0);
	m_filter = ri.Cvar_Get("m_filter", "0", 0);
	m_pitch = ri.Cvar_Get("m_pitch", "0.022", 0);
	m_yaw = ri.Cvar_Get("m_yaw", "0.022", 0);
	m_forward = ri.Cvar_Get("m_forward", "1", 0);
	m_side = ri.Cvar_Get("m_side", "0.8", 0);
	
	ri.Cmd_AddCommand("+mlook", REF_IN_MLookDown);
	ri.Cmd_AddCommand("-mlook", REF_IN_MLookUp);
	
	ri.Cmd_AddCommand("force_centerview", Force_CenterView_f);
	
	mx = my = 0.0;
	mouse_avail = true;
}

void REF_IN_Shutdown(void){
	if(mouse_avail){
		mouse_avail = false;
		
		ri.Cmd_RemoveCommand("+mlook");
		ri.Cmd_RemoveCommand("-mlook");
		
		ri.Cmd_RemoveCommand("force_centerview");
	}
}

/*
IN_Commands
*/
void REF_IN_Commands(void){
	int i;
	
	if(mouse_avail){
		for(i = 0; i < 3; i++){
			if((mouse_buttonstate &(1 << i)) && !(mouse_oldbuttonstate &(1 << i)))
				in_state->Key_Event_fp(K_MOUSE1 + i, true);
			if(!(mouse_buttonstate &(1 << i)) &&(mouse_oldbuttonstate &(1 << i)))
				in_state->Key_Event_fp(K_MOUSE1 + i, false);
		}
		if((mouse_buttonstate &(1 << 3)) && !(mouse_oldbuttonstate &(1 << 3)))
			in_state->Key_Event_fp(K_MOUSE4, true);
		if(!(mouse_buttonstate *(1 << 3)) &&(mouse_oldbuttonstate &(1 << 3)))
			in_state->Key_Event_fp(K_MOUSE4, false);
			
		if((mouse_buttonstate &(1 << 4)) && !(mouse_oldbuttonstate &(1 << 4)))
			in_state->Key_Event_fp(K_MOUSE5, true);
		if(!(mouse_buttonstate *(1 << 4)) &&(mouse_oldbuttonstate &(1 << 4)))
			in_state->Key_Event_fp(K_MOUSE5, false);
			
		mouse_oldbuttonstate = mouse_buttonstate;
	}
}

/*
IN_Move
*/
void REF_IN_Move(usercmd_t *cmd){
	
	if(mouse_avail){
		if(m_filter->value){
			mouse_x = (mx + old_mouse_x) * 0.5;
			mouse_y = (my + old_mouse_y) * 0.5;
		} else {
			mouse_x = mx;
			mouse_y = my;
		}
		
		old_mouse_x = mx;
		old_mouse_y = my;
		
		if(mouse_x || mouse_y){
			mouse_x *= sensitivity->value;
			mouse_y *= sensitivity->value;
			
			/* add mouse X/Y movement to cmd */
			if((*in_state->in_strafe_state & 1) || (lookstrafe->value && mlooking))
				cmd->sidemove += m_side->value * mouse_x;
			else in_state->viewangles[YAW] -= m_yaw->value * mouse_x;
				
			if((mlooking || freelook->value) && !(*in_state->in_strafe_state & 1))
				in_state->viewangles[PITCH] += m_pitch->value * mouse_y;
			else cmd->forwardmove -= m_forward->value * mouse_y;
				
			mx = my = 0;
		}
	}
}

static void IN_DeactivateMouse(void){
	if(!mouse_avail)
		return;
		
	if(mouse_active){
		mouse_active = false;
	}
}

static void IN_ActivateMouse(void){
	if(!mouse_avail)
		return;
		
	if(!mouse_active){
		mx = my = 0; // don't spazz
		mouse_active = true;
	}
}

void REF_IN_Frame(void){}

void REF_IN_Activate(qboolean active){
	if(active)
		IN_ActivateMouse();
	else
		IN_DeactivateMouse();
}

int XLateKey(SDL_Event *event){
	unsigned int keysym = event->key.keysym.sym;
	int key = 0;
	
	switch(keysym){
		case SDLK_KP9:
			key = K_KP_PGUP;
			break;
		case SDLK_PAGEUP:
			key = K_PGUP;
			break;
			
		case SDLK_KP3:
			key = K_KP_PGDN;
			break;
		case SDLK_PAGEDOWN:
			key = K_PGDN;
			break;
			
		case SDLK_KP7:
			key = K_KP_HOME;
			break;
		case SDLK_HOME:
			key = K_HOME;
			break;
			
		case SDLK_KP1:
			key = K_KP_END;
			break;
		case SDLK_END:
			key = K_END;
			break;
			
		case SDLK_KP4:
			key = K_KP_LEFTARROW;
			break;
		case SDLK_LEFT:
			key = K_LEFTARROW;
			break;
			
		case SDLK_KP6:
			key = K_KP_RIGHTARROW;
			break;
		case SDLK_RIGHT:
			key = K_RIGHTARROW;
			break;
			
		case SDLK_KP2:
			key = K_KP_DOWNARROW;
			break;
		case SDLK_DOWN:
			key = K_DOWNARROW;
			break;
			
		case SDLK_KP8:
			key = K_KP_UPARROW;
			break;
		case SDLK_UP:
			key = K_UPARROW;
			break;
			
		case SDLK_ESCAPE:
			key = K_ESCAPE;
			break;
			
		case SDLK_KP_ENTER:
			key = K_KP_ENTER;
			break;
		case SDLK_RETURN:
			key = K_ENTER;
			break;
			
		case SDLK_TAB:
			key = K_TAB;
			break;
			
		case SDLK_F1:
			key = K_F1;
			break;
		case SDLK_F2:
			key = K_F2;
			break;
		case SDLK_F3:
			key = K_F3;
			break;
		case SDLK_F4:
			key = K_F4;
			break;
		case SDLK_F5:
			key = K_F5;
			break;
		case SDLK_F6:
			key = K_F6;
			break;
		case SDLK_F7:
			key = K_F7;
			break;
		case SDLK_F8:
			key = K_F8;
			break;
		case SDLK_F9:
			key = K_F9;
			break;
			
		case SDLK_F10:
			if(event->type == SDL_KEYDOWN){
				if(vid_fullscreen->value){  //toggle fullscreen
					ri.Cvar_SetValue("vid_fullscreen", 0);
					ri.Cvar_SetValue("vid_width", vid_width->value / 2);
					ri.Cvar_SetValue("vid_height", vid_height->value / 2);
				}
				else {
					ri.Cvar_SetValue("vid_fullscreen", 1);
					ri.Cvar_SetValue("vid_width", vid_width->value * 2);
					ri.Cvar_SetValue("vid_height", vid_height->value * 2);
				}
			}
			
			ri.Cmd_ExecuteText(EXEC_NOW, "vid_restart");
			break;
			
		case SDLK_F11:
			if(event->type == SDL_KEYDOWN){
				if(_windowed_mouse->value)  //and windowed mouse
					ri.Cvar_SetValue("_windowed_mouse", 0);
				else 	ri.Cvar_SetValue("_windowed_mouse", 1);
			}
			break;
			
		case SDLK_F12:
			key = K_F12;
			break;
			
		case SDLK_BACKSPACE:
			key = K_BACKSPACE;
			break;
			
		case SDLK_KP_PERIOD:
			key = K_KP_DEL;
			break;
		case SDLK_DELETE:
			key = K_DEL;
			break;
			
		case SDLK_PAUSE:
			key = K_PAUSE;
			break;
			
		case SDLK_LSHIFT:
		case SDLK_RSHIFT:
			key = K_SHIFT;
			break;
			
		case SDLK_LCTRL:
		case SDLK_RCTRL:
			key = K_CTRL;
			break;
			
		case SDLK_LMETA:
		case SDLK_RMETA:
		case SDLK_LALT:
		case SDLK_RALT:
			key = K_ALT;
			break;
			
		case SDLK_KP5:
			key = K_KP_5;
			break;
			
		case SDLK_INSERT:
			key = K_INS;
			break;
		case SDLK_KP0:
			key = K_KP_INS;
			break;
			
		case SDLK_KP_MULTIPLY:
			key = '*';
			break;
		case SDLK_KP_PLUS:
			key = K_KP_PLUS;
			break;
		case SDLK_KP_MINUS:
			key = K_KP_MINUS;
			break;
		case SDLK_KP_DIVIDE:
			key = K_KP_SLASH;
			break;
			
		case SDLK_WORLD_7:
			key = '`';
			break;
			
		default:
			if(keysym < 128)
				key = keysym;
			break;
	}
	
	if(old_windowed_mouse != _windowed_mouse->value){
		old_windowed_mouse = _windowed_mouse->value;
		
		if(!_windowed_mouse->value){
			SDL_WM_GrabInput(SDL_GRAB_OFF);
		} else {
			SDL_WM_GrabInput(SDL_GRAB_ON);
		}
	}
	
	return key;
}

static unsigned char KeyStates[SDLK_LAST];

static void HandleEvents(SDL_Event *event){
	unsigned int key;
	
	switch(event->type){
		case SDL_MOUSEBUTTONDOWN:
			if(event->button.button == 4){
				keyq[keyq_head].key = K_MWHEELUP;
				keyq[keyq_head].down = true;
				keyq_head = (keyq_head + 1) & 63;
				keyq[keyq_head].key = K_MWHEELUP;
				keyq[keyq_head].down = false;
				keyq_head = (keyq_head + 1) & 63;
			} else if(event->button.button == 5){
				keyq[keyq_head].key = K_MWHEELDOWN;
				keyq[keyq_head].down = true;
				keyq_head = (keyq_head + 1) & 63;
				keyq[keyq_head].key = K_MWHEELDOWN;
				keyq[keyq_head].down = false;
				keyq_head = (keyq_head + 1) & 63;
			}
			break;
		case SDL_KEYDOWN:
			KeyStates[event->key.keysym.sym] = 1;
			key = XLateKey(event);
			if(key){
				keyq[keyq_head].key = key;
				keyq[keyq_head].down = true;
				keyq_head = (keyq_head + 1) & 63;
			}
			break;
		case SDL_KEYUP:
			if(KeyStates[event->key.keysym.sym]){
				KeyStates[event->key.keysym.sym] = 0;
				
				key = XLateKey(event);
				if(key){
					keyq[keyq_head].key = key;
					keyq[keyq_head].down = false;
					keyq_head =(keyq_head + 1) & 63;
				}
			}
			break;
		case SDL_QUIT:
			ri.Cmd_ExecuteText(EXEC_NOW, "quit");
			break;
	}
	
}

/*
** SWimp_Init
**
** This routine is responsible for initializing the implementation
** specific stuff in a software rendering subsystem.
*/
int SWimp_Init(void *hInstance, void *wndProc){
	if(SDL_WasInit(SDL_INIT_AUDIO | SDL_INIT_VIDEO) == 0){
		if(SDL_Init(SDL_INIT_VIDEO) < 0){
			Sys_Error("SDL Init failed: %s\n", SDL_GetError());
			return false;
		}
	} else if(SDL_WasInit(SDL_INIT_VIDEO) == 0){
		if(SDL_InitSubSystem(SDL_INIT_VIDEO) < 0){
			Sys_Error("SDL Init failed: %s\n", SDL_GetError());
			return false;
		}
	}
	return true;
}

#ifdef OPENGL
void *GLimp_GetProcAddress(const char *func){
	return SDL_GL_GetProcAddress(func);
}

int GLimp_Init(void *hInstance, void *wndProc){
	return SWimp_Init(hInstance, wndProc);
}
#endif

static void SetSDLIcon(void){
#include "icon.xbm"
	SDL_Surface * icon;
	SDL_Color color;
	Uint8 * ptr;
	int i, mask;
	
	icon = SDL_CreateRGBSurface(
			SDL_SWSURFACE, icon_width, icon_height, 
			8, 0, 0, 0, 0
	);
	
	if(icon == NULL)
		return;
	
	SDL_SetColorKey(icon, SDL_SRCCOLORKEY, 0);
	
	color.r = color.g = color.b = 255;
	SDL_SetColors(icon, &color, 0, 1);
	
	color.r = 0; color.g = 16; color.b = 0;
	SDL_SetColors(icon, &color, 1, 1);
	
	ptr = (Uint8 *)icon->pixels;
	for(i = 0; i < sizeof(icon_bits); i++){
		for(mask = 1; mask != 0x100; mask <<= 1){
			*ptr = (icon_bits[i] & mask) ? 1 : 0;
			ptr++;
		}
	}
	
	SDL_WM_SetIcon(icon, NULL);
	SDL_FreeSurface(icon);
}

/*
** SWimp_InitGraphics
**
** This initializes the software refresh's implementation specific
** graphics subsystem.  In the case of Windows it creates DIB or
** DDRAW surfaces.
**
** The necessary width and height parameters are grabbed from
** viddef.width and viddef.height.
*/
#ifndef OPENGL
static qboolean SWimp_InitGraphics(int width, int height, qboolean fullscreen){
	const SDL_VideoInfo *vinfo;
	int flags;
	
	if(surface && (surface->w == width) && (surface->h == height)){  //toggle fs
		
		int isfullscreen = (surface->flags & SDL_FULLSCREEN) ? 1 : 0;
		if(fullscreen != isfullscreen) SDL_WM_ToggleFullScreen(surface);
		
		isfullscreen = (surface->flags & SDL_FULLSCREEN) ? 1 : 0;
		if(fullscreen == isfullscreen) return true;
	}
	
	srandom(getpid());
	
	// free resources in use
	if(surface) SDL_FreeSurface(surface);
		
	// let the sound and input subsystems know about the new window
	ri.Vid_NewWindow(width, height);
	
	vinfo = SDL_GetVideoInfo();
	sdl_palettemode = (vinfo->vfmt->BitsPerPixel == 8) ?
			(SDL_PHYSPAL | SDL_LOGPAL) : SDL_LOGPAL;
	
	flags =  SDL_SWSURFACE | SDL_HWPALETTE;
	if(fullscreen) flags |= SDL_FULLSCREEN;
	
	if((surface = SDL_SetVideoMode(width, height, 8, flags)) == NULL){
		Sys_Error("(SOFTSDL) SDL SetVideoMode failed: %s\n", SDL_GetError());
		return false;
	}
	
	SDL_WM_SetCaption("Quetoo", "Quetoo");
	SDL_ShowCursor(0);
	SetSDLIcon();
	
	viddef.rowbytes = surface->pitch;
	viddef.buffer = surface->pixels;
	
	return true;
}
#else
static qboolean GLimp_InitGraphics(int width, int height, qboolean fullscreen){
	int flags;
	
	if(surface && (surface->w == width) && (surface->h == height)){  //toggle fs
		
		int isfullscreen = (surface->flags & SDL_FULLSCREEN) ? 1 : 0;
		if(fullscreen != isfullscreen) SDL_WM_ToggleFullScreen(surface);
		
		isfullscreen = (surface->flags & SDL_FULLSCREEN) ? 1 : 0;
		if(fullscreen == isfullscreen) return true;
	}
	
	srandom(getpid());
	
	// free resources in use
	if(surface) SDL_FreeSurface(surface);
	
	// let the sound and input subsystems know about the new window
	ri.Vid_NewWindow(width, height);
	
	SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	
	flags = SDL_OPENGL;
	if(fullscreen) flags |= SDL_FULLSCREEN;
	
	if((surface = SDL_SetVideoMode(width, height, 0, flags)) == NULL){
		Sys_Error("(SDLGL) SDL SetVideoMode failed: %s\n", SDL_GetError());
		return false;
	}
	
	SDL_WM_SetCaption("Quetoo", "Quetoo");
	SDL_ShowCursor(0);
	SetSDLIcon();
	
	return true;
}
#endif

#ifdef OPENGL
void GLimp_BeginFrame(){}
#endif

/*
** SWimp_EndFrame
**
** This does an implementation specific copy from the backbuffer to the
** front buffer.  In the Win32 case it uses BitBlt or BltFast depending
** on whether we're using DIB sections/GDI or DDRAW.
*/

#ifndef OPENGL
void SWimp_EndFrame(void){
	SDL_UpdateRect(surface, 0, 0, 0, 0);
}
#else
void GLimp_EndFrame(void){
	SDL_GL_SwapBuffers();
}
#endif

/*
* SWimp_SetMode
*/
#ifndef OPENGL
int SWimp_SetMode(int width, int height, qboolean fullscreen){
	
	if(!SWimp_InitGraphics(width, height, fullscreen)){
		return false;
	}
	
	R_GammaCorrectAndSetPalette((const unsigned char *) d_8to24table);	
	return true;
}
#else
int GLimp_SetMode(int width, int height, qboolean fullscreen){
	return GLimp_InitGraphics(width, height, fullscreen);
}
#endif

/*
** SWimp_SetPalette
**
** System specific palette setting routine.  A NULL palette means
** to use the existing palette.  The palette is expected to be in
** a padded 4-byte xRGB format.
*/
#ifndef OPENGL
void SWimp_SetPalette(const unsigned char *palette){
	SDL_Color colors[256];
	
	int i;
	
	if(!surface)
		return;
		
	if(!palette)
		palette = (const unsigned char *) sw_state.currentpalette;
		
	for(i = 0; i < 256; i++){
		colors[i].r = palette[i * 4 + 0];
		colors[i].g = palette[i * 4 + 1];
		colors[i].b = palette[i * 4 + 2];
	}
	
	SDL_SetPalette(surface, sdl_palettemode, colors, 0, 256);
}
#endif

/*
** SWimp_Shutdown
**
** System specific graphics subsystem shutdown routine.  Destroys
** DIBs or DDRAW surfaces as appropriate.
*/
void SWimp_Shutdown(void){
	if(surface)
		SDL_FreeSurface(surface);
	
	surface = NULL;
	
	if(SDL_WasInit(SDL_INIT_EVERYTHING) == SDL_INIT_VIDEO)
		SDL_Quit();
	else SDL_QuitSubSystem(SDL_INIT_VIDEO);
}

#ifdef OPENGL
void GLimp_Shutdown(void){
	SWimp_Shutdown();
}
#endif

/*
** SWimp_AppActivate
*/
#ifndef OPENGL
void SWimp_AppActivate(qboolean active){}
#else
void GLimp_AppActivate(qboolean active){}
#endif

Key_Event_fp_t Key_Event_fp;

void KBD_Init(Key_Event_fp_t fp){
	Key_Event_fp = fp;
}

void KBD_Update(void){
	static int KBD_Update_Flag;
	SDL_Event event;
	
	if(KBD_Update_Flag == 1)
		return;
		
	KBD_Update_Flag = 1;
	
	// get events from os
	if(surface){
		int bstate;
		
		while(SDL_PollEvent(&event))
			HandleEvents(&event);
			
		if(!mx && !my)
			SDL_GetRelativeMouseState(&mx, &my);
			
		mouse_buttonstate = 0;
		bstate = SDL_GetMouseState(NULL, NULL);
		if(SDL_BUTTON(1) & bstate)
			mouse_buttonstate |=(1 << 0);
		if(SDL_BUTTON(3) & bstate) /* quake2 has the right button be mouse2 */
			mouse_buttonstate |=(1 << 1);
		if(SDL_BUTTON(2) & bstate) /* quake2 has the middle button be mouse3 */
			mouse_buttonstate |=(1 << 2);
		if(SDL_BUTTON(6) & bstate)
			mouse_buttonstate |=(1 << 3);
		if(SDL_BUTTON(7) & bstate)
			mouse_buttonstate |=(1 << 4);
		
		while(keyq_head != keyq_tail){
			Key_Event_fp(keyq[keyq_tail].key, keyq[keyq_tail].down);
			keyq_tail =(keyq_tail + 1) & 63;
		}
	}
	
	KBD_Update_Flag = 0;
}

void KBD_Close(void){
	keyq_head = 0;
	keyq_tail = 0;
	
	memset(keyq, 0, sizeof(keyq));
}


#ifdef OPENGL
void Fake_glColorTableEXT(GLenum target, GLenum internalformat,
						   GLsizei width, GLenum format, GLenum type,
						   const GLvoid *table){
	byte temptable[256][4];
	byte *intbl;
	int i;
	
	for(intbl =(byte *)table, i = 0; i < 256; i++){
		temptable[i][2] = *intbl++;
		temptable[i][1] = *intbl++;
		temptable[i][0] = *intbl++;
		temptable[i][3] = 255;
	}
	qgl3DfxSetPaletteEXT((GLuint *)temptable);
}
#endif
