/*
* Copyright(c) 1997-2001 Id Software, Inc.
* Copyright(c) 2002 The Quakeforge Project.
* Copyright(c) 2006 Quetoo.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or(at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
* 
* See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/*
* This file contains ALL Linux specific stuff having to do with the
* software refresh.  When a port is being made the following functions
* must be implemented by the port:
*
* SWimp_EndFrame
* SWimp_Init
* SWimp_InitGraphics
* SWimp_SetPalette
* SWimp_Shutdown
*/

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <unistd.h>
#include <signal.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>

#ifdef HAVE_XF86_VIDMODE
#include <X11/extensions/xf86vmode.h>
#endif

#include "r_local.h"
#include "keys.h"

static GC	gc;
static Display	*dpy;
static int screen;
static Window win;
static XVisualInfo *visinfo;
static Atom wmDeleteWindow;
static int window_width;
static int window_height;

#define KEY_MASK (KeyPressMask | KeyReleaseMask)
#define MOUSE_MASK (ButtonPressMask | ButtonReleaseMask | PointerMotionMask | ButtonMotionMask)
#define X_MASK (KEY_MASK | MOUSE_MASK | VisibilityChangeMask | StructureNotifyMask)

// mouse vars
static qboolean mouse_avail;
static int mouse_buttonstate;
static int mouse_oldbuttonstate;
static int mouse_x, mouse_y;
static int mx, my;
static int old_mouse_x, old_mouse_y;
static int old_windowed_mouse;
static int win_x, win_y;

static qboolean mlooking;
static qboolean mouse_active = false;
static qboolean vidmode_active = false;

#ifdef HAVE_XF86_VIDMODE
static XF86VidModeModeInfo **vidmodes;
#endif

// state struct passed in Init
static in_state_t *in_state;

static cvar_t *in_mouse;
static cvar_t *_windowed_mouse;
static cvar_t *freelook;
static cvar_t *lookstrafe;
static cvar_t *sensitivity;
static cvar_t *m_filter;
static cvar_t *m_side;
static cvar_t *m_yaw;
static cvar_t *m_pitch;
static cvar_t *m_forward;

static Time myxtime;

static XImage *x_framebuffer = NULL;

typedef unsigned long PIXEL24;
static PIXEL24 st2d_8to24table[256];

static int shiftmask_fl = 0;
static long r_shift, g_shift, b_shift;
static unsigned long r_mask, g_mask, b_mask;

void shiftmask_init(void){
	unsigned int x;
	
	r_mask = visinfo->visual->red_mask;
	for(r_shift = -8, x = 1; x < r_mask; x = x << 1)
		r_shift++;
	
	g_mask = visinfo->visual->green_mask;
	for(g_shift = -8, x = 1; x < g_mask; x = x << 1)
		g_shift++;
	
	b_mask = visinfo->visual->blue_mask;
	for(b_shift = -8, x = 1;x < b_mask; x = x << 1)
		b_shift++;
	
	shiftmask_fl = 1;
}

PIXEL24 xlib_rgb24(int r, int g, int b){
	PIXEL24 p;
	if(shiftmask_fl == 0)
		shiftmask_init();
	p = 0;
	
	if(r_shift > 0){
		p =(r <<(r_shift)) & r_mask;
	} else if(r_shift < 0){
		p =(r >>(-r_shift)) & r_mask;
	} else
		p |=(r & r_mask);
		
	if(g_shift > 0){
		p |=(g <<(g_shift)) & g_mask;
	} else if(g_shift < 0){
		p |=(g >>(-g_shift)) & g_mask;
	} else
		p |=(g & g_mask);
		
	if(b_shift > 0){
		p |=(b <<(b_shift)) & b_mask;
	} else if(b_shift < 0){
		p |=(b >>(-b_shift)) & b_mask;
	} else
		p |=(b & b_mask);
		
	return p;
}


void st3_fixup(XImage *framebuf, int x, int y, int width, int height){
	int yi;
	byte *src;
	PIXEL24 *dest;
	register int count, n;
	
	if((x < 0) ||(y < 0))
		return;
		
	for(yi = y; yi < (y + height); yi++){
		src = (byte *) & framebuf->data [yi * framebuf->bytes_per_line];
		
		count = width;
		n = (count + 7) / 8;
		dest = ((PIXEL24 *)src) + x + width - 1;
		src += x + width - 1;
		
		switch(count % 8){
			case 0:
				do {
					*dest-- = st2d_8to24table[*src--];
				case 7:
					*dest-- = st2d_8to24table[*src--];
				case 6:
					*dest-- = st2d_8to24table[*src--];
				case 5:
					*dest-- = st2d_8to24table[*src--];
				case 4:
					*dest-- = st2d_8to24table[*src--];
				case 3:
					*dest-- = st2d_8to24table[*src--];
				case 2:
					*dest-- = st2d_8to24table[*src--];
				case 1:
					*dest-- = st2d_8to24table[*src--];
				} while(--n > 0);
		}
	}
}

static Cursor CreateNullCursor(Display *display, Window root){
	Pixmap cursormask;
	XGCValues xgc;
	GC gc;
	XColor dummycolour;
	Cursor cursor;
	
	cursormask = XCreatePixmap(display, root, 1, 1, 1);
	xgc.function = GXclear;
	gc = XCreateGC(display, cursormask, GCFunction, &xgc);
	XFillRectangle(display, cursormask, gc, 0, 0, 1, 1);
	dummycolour.pixel = 0;
	dummycolour.red = 0;
	dummycolour.flags = 4;
	cursor = XCreatePixmapCursor(display, cursormask, cursormask,
								 &dummycolour, &dummycolour, 0, 0);
	XFreePixmap(display, cursormask);
	XFreeGC(display, gc);
	return cursor;
}

static void install_grabs(void){
	
	XDefineCursor(dpy, win, CreateNullCursor(dpy, win));
	
	XGrabPointer(dpy, win,
				True,
				0,
				GrabModeAsync, GrabModeAsync,
				win,
				None,
				CurrentTime);
				
	XWarpPointer(dpy, None, win,
				0, 0, 0, 0,
				viddef.width / 2, viddef.height / 2);
				
	XGrabKeyboard(dpy, win,
				False,
				GrabModeAsync, GrabModeAsync,
				CurrentTime);
				
	mouse_active = true;
}

static void uninstall_grabs(void){
	if(!dpy || !win)
		return;
	
	XUngrabPointer(dpy, CurrentTime);
	XUngrabKeyboard(dpy, CurrentTime);
	XUndefineCursor(dpy, win);
	
	mouse_active = false;
}

static void Force_CenterView_f(void){
	in_state->viewangles[PITCH] = 0;
}

static void REF_IN_MLookDown(void){
	mlooking = true;
}

static void REF_IN_MLookUp(void){
	mlooking = false;
	in_state->IN_CenterView_fp();
}

void REF_IN_Init(in_state_t *in_state_p){
	in_state = in_state_p;
	
	in_mouse = ri.Cvar_Get("in_mouse", "1", CVAR_ARCHIVE);
	_windowed_mouse = ri.Cvar_Get("_windowed_mouse", "1", CVAR_ARCHIVE);
	freelook = ri.Cvar_Get("freelook", "0", 0);
	lookstrafe = ri.Cvar_Get("lookstrafe", "0", 0);
	sensitivity = ri.Cvar_Get("sensitivity", "3", 0);
	m_filter = ri.Cvar_Get("m_filter", "0", 0);
	m_pitch = ri.Cvar_Get("m_pitch", "0.022", 0);
	m_yaw = ri.Cvar_Get("m_yaw", "0.022", 0);
	m_forward = ri.Cvar_Get("m_forward", "1", 0);
	m_side = ri.Cvar_Get("m_side", "0.8", 0);
	
	ri.Cmd_AddCommand("+mlook", REF_IN_MLookDown);
	ri.Cmd_AddCommand("-mlook", REF_IN_MLookUp);
	
	ri.Cmd_AddCommand("force_centerview", Force_CenterView_f);
	
	mx = my = 0.0;
	mouse_avail = true;
}

void REF_IN_Shutdown(void){
	if(mouse_avail){
		mouse_avail = false;
		
		ri.Cmd_RemoveCommand("+mlook");
		ri.Cmd_RemoveCommand("-mlook");
		ri.Cmd_RemoveCommand("force_centerview");
	}
}

/*
IN_Commands
*/
void REF_IN_Commands(void){
	int i;
	
	if(mouse_avail){
		for(i = 0; i < 3; i++){
			if((mouse_buttonstate &(1 << i)) && !(mouse_oldbuttonstate &(1 << i)))
				in_state->Key_Event_fp(K_MOUSE1 + i, true);
			if(!(mouse_buttonstate &(1 << i)) &&(mouse_oldbuttonstate &(1 << i)))
				in_state->Key_Event_fp(K_MOUSE1 + i, false);
		}
		if((mouse_buttonstate &(1 << 3)) && !(mouse_oldbuttonstate &(1 << 3)))
			in_state->Key_Event_fp(K_MOUSE4, true);
		if(!(mouse_buttonstate *(1 << 3)) &&(mouse_oldbuttonstate &(1 << 3)))
			in_state->Key_Event_fp(K_MOUSE4, false);
			
		if((mouse_buttonstate &(1 << 4)) && !(mouse_oldbuttonstate &(1 << 4)))
			in_state->Key_Event_fp(K_MOUSE5, true);
		if(!(mouse_buttonstate *(1 << 4)) &&(mouse_oldbuttonstate &(1 << 4)))
			in_state->Key_Event_fp(K_MOUSE5, false);
			
		mouse_oldbuttonstate = mouse_buttonstate;
	}
}

/*
IN_Move
*/
void REF_IN_Move(usercmd_t *cmd){
	
	if(mouse_avail){
		if(m_filter->value){
			mouse_x = (mx + old_mouse_x) * 0.5;
			mouse_y = (my + old_mouse_y) * 0.5;
		} else {
			mouse_x = mx;
			mouse_y = my;
		}
		
		old_mouse_x = mx;
		old_mouse_y = my;
		
		if(mouse_x || mouse_y){
			mouse_x *= sensitivity->value;
			mouse_y *= sensitivity->value;
			
			/* add mouse X/Y movement to cmd */
			if((*in_state->in_strafe_state & 1) || (lookstrafe->value && mlooking))
				cmd->sidemove += m_side->value * mouse_x;
			else in_state->viewangles[YAW] -= m_yaw->value * mouse_x;
			
			if((mlooking || freelook->value) && !(*in_state->in_strafe_state & 1))
				in_state->viewangles[PITCH] += m_pitch->value * mouse_y;
			else cmd->forwardmove -= m_forward->value * mouse_y;
			
			mx = my = 0;
		}
	}
}

static void IN_DeactivateMouse(void){
	if(!mouse_avail || !dpy || !win)
		return;
		
	if(mouse_active){
		uninstall_grabs();
		mouse_active = false;
	}
}

static void IN_ActivateMouse(void){
	if(!mouse_avail || !dpy || !win)
		return;
	
	if(!mouse_active){
		mx = my = 0; // don't spazz
		install_grabs();
		mouse_active = true;
	}
}

void REF_IN_Frame(void){
	int i;
	
	if(mouse_avail){
		for(i = 0; i < 3; i++){
			if((mouse_buttonstate & (1 << i)) && !(mouse_oldbuttonstate & (1 << i)))
				in_state->Key_Event_fp(K_MOUSE1 + i, true);
				
			if(!(mouse_buttonstate & (1 << i)) && (mouse_oldbuttonstate & (1 << i)))
				in_state->Key_Event_fp(K_MOUSE1 + i, false);
		}
		mouse_oldbuttonstate = mouse_buttonstate;
	}
}

void REF_IN_Activate(qboolean active){
	if(active || vidmode_active)
		IN_ActivateMouse();
	else IN_DeactivateMouse();
}

/*
* XLateKey
*/
static int XLateKey(XKeyEvent *ev){

	int key;
	char buf[64];
	KeySym keysym;
	
	key = 0;
	
	XLookupString(ev, buf, sizeof buf, &keysym, 0);
	
	switch(keysym){
		case XK_KP_Page_Up:
			key = K_KP_PGUP;
			break;
		case XK_Page_Up:
			key = K_PGUP;
			break;
			
		case XK_KP_Page_Down:
			key = K_KP_PGDN;
			break;
		case XK_Page_Down:
			key = K_PGDN;
			break;
			
		case XK_KP_Home:
			key = K_KP_HOME;
			break;
		case XK_Home:
			key = K_HOME;
			break;
			
		case XK_KP_End:
			key = K_KP_END;
			break;
		case XK_End:
			key = K_END;
			break;
			
		case XK_KP_Left:
			key = K_KP_LEFTARROW;
			break;
		case XK_Left:
			key = K_LEFTARROW;
			break;
			
		case XK_KP_Right:
			key = K_KP_RIGHTARROW;
			break;
		case XK_Right:
			key = K_RIGHTARROW;
			break;
			
		case XK_KP_Down:
			key = K_KP_DOWNARROW;
			break;
		case XK_Down:
			key = K_DOWNARROW;
			break;
			
		case XK_KP_Up:
			key = K_KP_UPARROW;
			break;
		case XK_Up:
			key = K_UPARROW;
			break;
			
		case XK_Escape:
			key = K_ESCAPE;
			break;
			
		case XK_KP_Enter:
			key = K_KP_ENTER;
			break;
		case XK_Return:
			key = K_ENTER;
			break;
			
		case XK_Tab:
			key = K_TAB;
			break;
			
		case XK_F1:
			key = K_F1;
			break;
			
		case XK_F2:
			key = K_F2;
			break;
			
		case XK_F3:
			key = K_F3;
			break;
			
		case XK_F4:
			key = K_F4;
			break;
			
		case XK_F5:
			key = K_F5;
			break;
			
		case XK_F6:
			key = K_F6;
			break;
			
		case XK_F7:
			key = K_F7;
			break;
			
		case XK_F8:
			key = K_F8;
			break;
			
		case XK_F9:
			key = K_F9;
			break;
			
		case XK_F10:
			if(ev->type == KeyPress){
				if(vid_fullscreen->value){  //toggle fullscreen
					ri.Cvar_SetValue("vid_fullscreen", 0);
					ri.Cvar_SetValue("vid_width", vid_width->value / 2);
					ri.Cvar_SetValue("vid_height", vid_height->value / 2);
				}
				else {
					ri.Cvar_SetValue("vid_fullscreen", 1);
					ri.Cvar_SetValue("vid_width", vid_width->value * 2);
					ri.Cvar_SetValue("vid_height", vid_height->value * 2);
				}
				
				ri.Cmd_ExecuteText(EXEC_NOW, "vid_restart");
			}
			break;
			
		case XK_F11:
			if(ev->type == KeyPress){
				if(_windowed_mouse->value)  //and windowed mouse
					ri.Cvar_SetValue("_windowed_mouse", 0);
				else 	ri.Cvar_SetValue("_windowed_mouse", 1);
			}
			break;
			
		case XK_F12:
			key = K_F12;
			break;
			
		case XK_BackSpace:
			key = K_BACKSPACE;
			break;
			
		case XK_KP_Delete:
			key = K_KP_DEL;
			break;
		case XK_Delete:
			key = K_DEL;
			break;
			
		case XK_Pause:
			key = K_PAUSE;
			break;
			
		case XK_Shift_L:
		case XK_Shift_R:
			key = K_SHIFT;
			break;
			
		case XK_Execute:
		case XK_Control_L:
		case XK_Control_R:
			key = K_CTRL;
			break;
			
		case XK_Alt_L:
		case XK_Meta_L:
		case XK_Alt_R:
		case XK_Meta_R:
			key = K_ALT;
			break;
			
		case XK_KP_Begin:
			key = K_KP_5;
			break;
			
		case XK_Insert:
			key = K_INS;
			break;
		case XK_KP_Insert:
			key = K_KP_INS;
			break;
			
		case XK_KP_Multiply:
			key = '*';
			break;
		case XK_KP_Add:
			key = K_KP_PLUS;
			break;
		case XK_KP_Subtract:
			key = K_KP_MINUS;
			break;
		case XK_KP_Divide:
			key = K_KP_SLASH;
			break;
		
		default:
			key = *(unsigned char*)buf;
			if(key >= 'A' && key <= 'Z')
				key = key - 'A' + 'a';
			if(key >= 1 && key <= 26) /* ctrl+alpha */
				key = key + 'a' - 1;
			break;
	}
	
	return key;
}


static void HandleEvents(void){
	XEvent event;
	int b;
	qboolean dowarp = false;
	int mwx = viddef.width / 2;
	int mwy = viddef.height / 2;
	
	if(!dpy)
		return;
		
	while(XPending(dpy)){
	
		XNextEvent(dpy, &event);
		
		switch(event.type){
			case KeyPress:
				myxtime = event.xkey.time;
			case KeyRelease:
				if(in_state && in_state->Key_Event_fp)
					in_state->Key_Event_fp(XLateKey(&event.xkey), event.type == KeyPress);
				break;
				
			case MotionNotify:
				if(mouse_active){
					if(_windowed_mouse->value){
						int xoffset = ((int)event.xmotion.x - (int)(window_width / 2));
						int yoffset = ((int)event.xmotion.y - (int)(window_height / 2));
						
						if(xoffset != 0 || yoffset != 0){
						
							mx += xoffset;
							my += yoffset;
							
							XSelectInput(dpy, win, X_MASK & ~PointerMotionMask);
							XWarpPointer(dpy, None, win, 0, 0, 0, 0,
										 (window_width / 2),(window_height / 2));
							XSelectInput(dpy, win, X_MASK);
						}
					} else {
						mx +=((int)event.xmotion.x - mwx) * 2;
						my +=((int)event.xmotion.y - mwy) * 2;
						mwx = event.xmotion.x;
						mwy = event.xmotion.y;
						
						if(mx || my)
							dowarp = true;
					}
				}
				break;
				
			case ButtonPress:
				myxtime = event.xbutton.time;
				
				b = -1;
				if(event.xbutton.button == 1)
					b = 0;
				else if(event.xbutton.button == 2)
					b = 2;
				else if(event.xbutton.button == 3)
					b = 1;
				else if(event.xbutton.button == 4)
					in_state->Key_Event_fp(K_MWHEELUP, 1);
				else if(event.xbutton.button == 5)
					in_state->Key_Event_fp(K_MWHEELDOWN, 1);
				if(b >= 0 && in_state && in_state->Key_Event_fp)
					in_state->Key_Event_fp(K_MOUSE1 + b, true);
				break;
				
			case ButtonRelease:
				b = -1;
				if(event.xbutton.button == 1)
					b = 0;
				else if(event.xbutton.button == 2)
					b = 2;
				else if(event.xbutton.button == 3)
					b = 1;
				else if(event.xbutton.button == 4)
					in_state->Key_Event_fp(K_MWHEELUP, 0);
				else if(event.xbutton.button == 5)
					in_state->Key_Event_fp(K_MWHEELDOWN, 0);
				if(b >= 0 && in_state && in_state->Key_Event_fp)
					in_state->Key_Event_fp(K_MOUSE1 + b, false);
				break;
				
			case CreateNotify:
				win_x = event.xcreatewindow.x;
				win_y = event.xcreatewindow.y;
				break;
				
			case ConfigureNotify:
				win_x = event.xconfigure.x;
				win_y = event.xconfigure.y;
				break;
				
			case ClientMessage:
				if(event.xclient.data.l[0] == wmDeleteWindow)
					ri.Cmd_ExecuteText(EXEC_NOW, "quit");
				break;
				
			case MapNotify:
				if(_windowed_mouse->value){
					XGrabPointer(dpy, win, True, 0, GrabModeAsync,
								  GrabModeAsync, win, None, CurrentTime);
				}
				break;
				
			case UnmapNotify:
				if(_windowed_mouse->value){
					XUngrabPointer(dpy, CurrentTime);
				}
				break;
		}
	}
	
	if(old_windowed_mouse != _windowed_mouse->value){
		old_windowed_mouse = _windowed_mouse->value;
		
		if(!_windowed_mouse->value){  //ungrab pointer
			XUngrabPointer(dpy, CurrentTime);
		} else {  //grab pointer
			XGrabPointer(dpy, win, True, 0, GrabModeAsync,
						  GrabModeAsync, win, None, CurrentTime);
		}
	}
	
	if(dowarp){  //move cursor to center
		XWarpPointer(dpy, None, win, 
				0, 0, 0, 0, viddef.width / 2, viddef.height / 2
		);
	}
}

Key_Event_fp_t Key_Event_fp;

void KBD_Init(Key_Event_fp_t fp){
	Key_Event_fp = fp;
}

void KBD_Update(void){
	HandleEvents();
}

void KBD_Close(void){}

char *REF_Sys_GetClipboardData(){
	Window sowner;
	Atom type, property;
	unsigned long len, bytes_left, tmp;
	unsigned char *data;
	int format, result;
	char *ret = NULL;
	
	sowner = XGetSelectionOwner(dpy, XA_PRIMARY);
	
	if(sowner != None){
		property = XInternAtom(dpy,
							   "GETCLIPBOARDDATA_PROP",
							   False);
		                       
		XConvertSelection(dpy,
						  XA_PRIMARY, XA_STRING,
						  property, win, myxtime);  // myxtime is time of last X event
		
		XFlush(dpy);
		
		XGetWindowProperty(dpy,
						   win, property,
						   0, 0, False, AnyPropertyType,
						   &type, &format, &len,
						   &bytes_left, &data);
						   
		if(bytes_left > 0){
			result = XGetWindowProperty(dpy,
								   win, property,
								   0, bytes_left, True, AnyPropertyType,
								   &type, &format, &len,
								   &tmp, &data);
								   
			if(result == Success){
				ret = strdup((char*) data);
			}
			XFree(data);
		}
	}
	return ret;
}

void signal_handler(int sig){
	printf("Received signal %d, exiting...\n", sig);
	SWimp_Shutdown();
	_exit(0);
}

/*
** SWimp_Init
**
** This routine is responsible for initializing the implementation
** specific stuff in a software rendering subsystem.
*/
int SWimp_Init(void *hInstance, void *wndProc){
	
	signal(SIGHUP, signal_handler);
	signal(SIGQUIT, signal_handler);
	signal(SIGILL, signal_handler);
	signal(SIGTRAP, signal_handler);
	signal(SIGIOT, signal_handler);
	signal(SIGBUS, signal_handler);
	signal(SIGFPE, signal_handler);
	signal(SIGSEGV, signal_handler);
	signal(SIGTERM, signal_handler);
	
	return true;
}

/*
** SWimp_SetMode
*/
int SWimp_SetMode(int width, int height, qboolean fullscreen){
	
	Window root;
	XSetWindowAttributes attr;
	XSizeHints *sizehints;
	XWMHints *wmhints;
	unsigned long mask;
	int i;
	
	window_width = width;
	window_height = height;
	
	// destroy existing window
	SWimp_Shutdown();
	
	if(!(dpy = XOpenDisplay(NULL))){
		fprintf(stderr, "Error couldn't open the X display\n");
		return false;
	}
	
	screen = DefaultScreen(dpy);
	root = RootWindow(dpy, screen);
	
	visinfo = (XVisualInfo *)malloc(sizeof(XVisualInfo));
	memset(visinfo, 0, sizeof(visinfo));
	
	if(!XMatchVisualInfo(dpy, screen, 24, DirectColor, visinfo)){
		fprintf(stderr, "E: couldn't get visual at 24 bit\n");
		return false;
	}
	
#ifdef HAVE_XF86_VIDMODE
	int num_vidmodes = 0;
	
	XF86VidModeGetAllModeLines(dpy, screen, &num_vidmodes, &vidmodes);
	
	if(fullscreen){
		int best_dist = 9999999;
		int best_fit = -1;
		int x, y, dist;
		
		for(i = 0; i < num_vidmodes; i++){
			if(width > vidmodes[i]->hdisplay ||
					height > vidmodes[i]->vdisplay)
				continue;
				
			x = width - vidmodes[i]->hdisplay;
			y = height - vidmodes[i]->vdisplay;
			
			dist = (x * x) + (y * y);
			if(dist < best_dist){
				best_dist = dist;
				best_fit = i;
			}
		}
		
		if(best_fit != -1){
			XF86VidModeSwitchToMode(dpy, screen, vidmodes[best_fit]);
			XF86VidModeSetViewPort(dpy, screen, 0, 0);
			
			width = vidmodes[best_fit]->hdisplay;
			height = vidmodes[best_fit]->vdisplay;
			
			vidmode_active = true;
		}
		else {
			vidmode_active = false;
			fullscreen = false;
		}
	}
#endif // HAVE_XF86_VIDMODE
	
	// window attributes
	attr.background_pixel = 0;
	attr.border_pixel = 0;
	attr.colormap = XCreateColormap(dpy, root, visinfo->visual, AllocNone);
	attr.event_mask = X_MASK;
	
	// setup attributes for main window
	if(vidmode_active){
		mask = CWBackPixel | CWColormap | CWSaveUnder | CWBackingStore |
				CWEventMask | CWOverrideRedirect;
		
		attr.override_redirect = True;
		attr.backing_store = NotUseful;
		attr.save_under = False;
	}
	else mask = CWBackPixel | CWBorderPixel | CWColormap | CWEventMask;
	
	// create window
	win = XCreateWindow(dpy, root, 0, 0, width, height,
			0, visinfo->depth, InputOutput,
			visinfo->visual, mask, &attr
	);
	
	sizehints = XAllocSizeHints();
	if(sizehints){
		sizehints->min_width = width;
		sizehints->min_height = height;
		sizehints->max_width = width;
		sizehints->max_height = height;
		sizehints->base_width = width;
		sizehints->base_height = height;
		
		sizehints->flags = PMinSize | PMaxSize | PBaseSize;
	}
	
	wmhints = XAllocWMHints();
	if(wmhints){
		#include "icon.xbm"
		
		Pixmap icon_pixmap, icon_mask;
		unsigned long fg, bg;
		int i;
		
		fg = BlackPixel(dpy, visinfo->screen);
		bg = WhitePixel(dpy, visinfo->screen);
		
		icon_pixmap = XCreatePixmapFromBitmapData(
				dpy, win, (char *)icon_bits, icon_width, icon_height,
				fg, bg, visinfo->depth
		);
		
		for(i = 0; i < sizeof(icon_bits); i++)
			icon_bits[i] = ~icon_bits[i];
		
		icon_mask = XCreatePixmapFromBitmapData(
				dpy, win,(char *)icon_bits, icon_width, icon_height,
				bg, fg, visinfo->depth
		);
		
		wmhints->flags = IconPixmapHint | IconMaskHint;
		wmhints->icon_pixmap = icon_pixmap;
		wmhints->icon_mask = icon_mask;
	}
	
	XSetWMProperties(dpy, win, NULL, NULL, NULL, 0,
			sizehints, wmhints, None
	);
	
	if(sizehints) XFree(sizehints);
	if(wmhints) XFree(wmhints);
	
	XStoreName(dpy, win, "Quetoo");
	
	wmDeleteWindow = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
	XSetWMProtocols(dpy, win, &wmDeleteWindow, 1);
	
	XMapWindow(dpy, win);
	
#ifdef HAVE_XF86_VIDMODE
	if(vidmode_active){
		XMoveWindow(dpy, win, 0, 0);
		XRaiseWindow(dpy, win);
		XWarpPointer(dpy, None, win, 0, 0, 0, 0, 0, 0);
	}
#endif // HAVE_XF86_VIDMODE
	
	XGCValues vals;
	vals.graphics_exposures = False;
	gc = XCreateGC(dpy, win, GCGraphicsExposures, &vals);
	
	R_GammaCorrectAndSetPalette((const unsigned char *)d_8to24table);
	
	if(x_framebuffer){
		free(x_framebuffer->data);
		free(x_framebuffer);
	}
	
	viddef.height = height; viddef.width = width;
	i = ((viddef.width * 4 + 7) & ~7) * viddef.height;
	
	x_framebuffer = XCreateImage(dpy,
			visinfo->visual,
			visinfo->depth,
			ZPixmap,
			0,
			malloc(i),
			viddef.width, viddef.height,
			16,
			0);
				
	if(!x_framebuffer)
		Sys_Error("XCreateImage failed\n");
	
	viddef.rowbytes = x_framebuffer->bytes_per_line;
	viddef.buffer = (byte *)x_framebuffer->data;
	
	ri.Vid_NewWindow(width, height);
	
	XDefineCursor(dpy, win, CreateNullCursor(dpy, win));
	
	return true;
}

/*
** SWimp_SetPalette
**
** System specific palette setting routine.  A NULL palette means
** to use the existing palette.  The palette is expected to be in
** a padded 4-byte xRGB format.
*/
void SWimp_SetPalette(const unsigned char *palette){
	int i;
	
	if(!dpy || !win)
		return;
		
	if(!palette)
		palette = (const unsigned char *)sw_state.currentpalette;
		
	for(i = 0; i < 256; i++){
		st2d_8to24table[i] = xlib_rgb24(
			palette[i * 4], palette[i * 4 + 1], palette[i * 4 + 2]
		);
	}
}

/*
** SWimp_BeginFrame
*/
void SWImp_BeginFrame(void){}

/*
** SWimp_EndFrame
**
** Copy from the back buffer to the front buffer.
*/
void SWimp_EndFrame(void){
	st3_fixup(x_framebuffer, 0, 0, viddef.width, viddef.height);
	XPutImage(dpy, win, gc, x_framebuffer, 0, 0, 0, 0, viddef.width, viddef.height);
	XSync(dpy, False);
}

/*
** SWimp_Shutdown
**
** System specific graphics subsystem shutdown routine.  Destroys
** DIBs or DDRAW surfaces as appropriate.
*/
void SWimp_Shutdown(void){
	uninstall_grabs();
	mouse_active = false;
	
	if(dpy){
		if(win)
			XDestroyWindow(dpy, win);

#ifdef HAVE_XF86_VIDMODE
		if(vidmode_active)
			XF86VidModeSwitchToMode(dpy, screen, vidmodes[0]);
#endif // HAVE_XF86_VIDMODE
		
		XUngrabKeyboard(dpy, CurrentTime);
		XCloseDisplay(dpy);
	}
	
	if(x_framebuffer){
		free(x_framebuffer->data);
		free(x_framebuffer);
		x_framebuffer = NULL;
	}
	
	dpy = NULL;
	win = 0;
}

/*
** SWimp_AppActivate
*/
void SWimp_AppActivate(qboolean active){}
